""" --- palma ---
	Some useful functions and classes used by palma.py
"""

__version__ = '0.3.3'

# Copyright notice string
__copyright__ = """\
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
Written (W) 2006 Uta Schulze
Copyright (C) 2006 Max-Planck-Society
"""


import sys
import numpy
import palma_utils as pu
import alignment_tool as at


class alignment_info: 
	nr_paths_found = -1
	score = []
	match = []
	mism = []
	qGapb = []
	tGapb = []
	qStart = []
	qEnd = []
	tStart = []
	tEnd = []
	num_exons = []
	qExonSizes = []
	qStarts = []
	qEnds = []
	tExonSizes = []
	tStarts = []
	tEnds = []
	exon_length = []
	identity = []
	comparison = []
	qIDX = []
	tIDX = []
	dna_letters = []
	est_letters = []
	
def do_alignment(dna, dna_len, est, est_len, reduce_region, model, t_strand,\
		 translation, comparison_char, don_support_values, acc_support_values):
	import time

	nr_paths = 1 
	mlen = 6 #length of match matrix, const

	align_info = alignment_info()
	align_info.score = []
	align_info.match = []
	align_info.mism = []
	align_info.qGapb = []
	align_info.tGapb = []
	align_info.qStart = []
	align_info.qEnd = []
	align_info.tStart = []
	align_info.tEnd = []
	align_info.num_exons = []
	align_info.qExonSizes = []
	align_info.qStarts = []
	align_info.qEnds = []
	align_info.tExonSizes = []
	align_info.tStarts = []
	align_info.tEnds = []
	align_info.exon_length = []
	align_info.identity = []
	align_info.comparison = []
	align_info.qIDX = []
	align_info.tIDX = []
	align_info.dna_letters = []
	align_info.est_letters = []
	align_info.splice_align = [] 

	t0 = time.time() 
	#<------------------------------------------------------------------------------>
	# assigns parameter [h(30),d(30),a(30),mmatrix(36)]
	# computes scores for donor/acceptor sites using p.l. function
	#<------------------------------------------------------------------------------>
	h,d,a,mmatrix = pu.set_param(model)
	donor	= don_support_values
	acceptor = acc_support_values
	for cnt in range(dna_len):
		if numpy.isfinite(don_support_values[cnt]):
			donor[cnt] = pu.penalty_lookup(d,don_support_values[cnt])
		if numpy.isfinite(acc_support_values[cnt]):
			acceptor[cnt] = pu.penalty_lookup(a,acc_support_values[cnt])
	del don_support_values
	del acc_support_values
	#<------------------------------------------------------------------------------>

	#print time.time() - t0
	
	#<------------------------------------------------------------------------------>
	# convert list to double-Array (py -> c++), 
	#<------------------------------------------------------------------------------>
	mmatrix_array  = at.createDoubleArrayFromList(mmatrix)
	donor_array	= at.createDoubleArrayFromList(donor)
	acceptor_array = at.createDoubleArrayFromList(acceptor)
	h_struct = at.penalty()
	h_struct.len = h.len 
	h_struct.limits = at.createDoubleArrayFromList(h.limits)
	h_struct.penalties = at.createDoubleArrayFromList(h.penalties)
	h_struct.name = h.name 
	h_struct.max_len = h.max_len 
	h_struct.min_len = h.min_len 
	h_struct.transform = h.transform 
	#<------------------------------------------------------------------------------>

	#print time.time() - t0

	#<------------------------------------------------------------------------------>
	# call myalign
	#<------------------------------------------------------------------------------>
	BLA = at.align_info() ; #not necessary
	try:
		BLA = at.myalign(reduce_region, mlen, dna_len, dna, est_len, est, h_struct,\
						 mmatrix_array, donor_array, acceptor_array)
	except MemoryError:
		raise(MemoryError)

	#print time.time() - t0

#<------------------------------------------------------------------------------>


#<------------------------------------------------------------------------------>
# myalign Output - convert arrays to list (c++ -> py)
#<------------------------------------------------------------------------------>
	splice_align_array  = at.createListFromIntArray(BLA.splice_align, dna_len*nr_paths)
	est_align_array	 = at.createListFromIntArray(BLA.est_align, est_len*nr_paths)
	mmatrix_param_array = at.createListFromIntArray(BLA.mmatrix_param, mlen*mlen*nr_paths)
	alignment_scores	= at.createListFromDoubleArray(BLA.alignment_scores,nr_paths)
	alignment_length	= at.createListFromIntArray(BLA.alignment_length, nr_paths)
	#print "before crash on intel mac"	
	aligned_dna_array   = at.createListFromIntArray(BLA.aligned_dna, alignment_length[0])
	aligned_est_array   = at.createListFromIntArray(BLA.aligned_est, alignment_length[0])
	#print "after crash on intel mac"	
#<------------------------------------------------------------------------------>

	#print alignment_scores

#<------------------------------------------------------------------------------>
# free memory
#<------------------------------------------------------------------------------>
	at.delete_intArray(BLA.splice_align)
	at.delete_intArray(BLA.est_align)
	at.delete_intArray(BLA.mmatrix_param)
	at.delete_intArray(BLA.aligned_dna)
	at.delete_intArray(BLA.aligned_est)
	at.delete_doubleArray(BLA.alignment_scores)
	at.delete_intArray(BLA.alignment_length)
    
	del BLA
	del h_struct
	at.delete_doubleArray(mmatrix_array)
	at.delete_doubleArray(donor_array)
	at.delete_doubleArray(acceptor_array)
#<------------------------------------------------------------------------------>


#<------------------------------------------------------------------------------>
# myalign output: creating lists of information
# i.th entry in list is alignment for i.th path (nr_path many)
#<------------------------------------------------------------------------------>
	mmatrix_param = []
	splice_align  = []
	est_align	 = []
	dna_numbers   = []
	est_numbers   = []
	dna_letters   = []
	est_letters   = []

	for i in range(nr_paths):
		result_length = alignment_length[i] #length of i.th alignment
		if result_length == 0:
			#no alignment found
			nr_paths = i-1+1 #+1 because index starts with zero
			alignment_scores = alignment_scores[0:nr_paths]
			alignment_length = alignment_length[0:nr_paths]
			assert(len(mmatrix_param)==nr_paths)
			assert(len(splice_align)==nr_paths)
			assert(len(est_align)==nr_paths)
			assert(len(dna_numbers)==nr_paths)
			assert(len(est_numbers)==nr_paths)
			assert(len(dna_letters)==nr_paths)
			assert(len(est_letters)==nr_paths)
			break

		#match matrix for the i.th path (numbers of matches, mismatches and gaps):
		mmatrix_param.append(mmatrix_param_array[i*mlen*mlen:(i+1)*mlen*mlen])
	
		#splice_align and est_align for the i.th path (show dangling ends and exons/introns):
		splice_align.append(splice_align_array[i*dna_len:(i+1)*dna_len])
		est_align.append(est_align_array[i*est_len:(i+1)*est_len])		
	
	dna_numbers.append(aligned_dna_array[0:result_length]) #alignment in numbers (see translation)
	est_numbers.append(aligned_est_array[0:result_length]) #alignment in numbers (see translation)
	aligned_dna_array = aligned_dna_array[result_length:]  #shorten aligned_dna_array
	aligned_est_array = aligned_est_array[result_length:]  #shorten aligned_est_array
	
	dna_letters.append(reduce(lambda x,y: x+y, map(lambda x: translation[int(x)],dna_numbers[i])))
	est_letters.append(reduce(lambda x,y: x+y, map(lambda x: translation[int(x)],est_numbers[i])))
	
	if nr_paths==0:
		#absolutely no alignment found
		align_info.nr_paths_found = nr_paths
		return align_info

	assert(len(aligned_dna_array)==0)
	assert(len(aligned_est_array)==0)
#<------------------------------------------------------------------------------>


#<------------------------------------------------------------------------------>
# analysing alignments (n-best)
#<------------------------------------------------------------------------------>
	if 0:
		print alignment_scores
		print alignment_length
		print mmatrix_param
		print splice_align
		print est_align
		print dna_numbers
		print est_numbers
		print dna_letters
		print est_letters 

	align_info.nr_paths_found = nr_paths

	ai = [] ;
	for i in range(nr_paths):
		(match, mism, qGapb, tGapb) = pu.evaluate_mmatrix(mmatrix_param[i])
		if (match +mism +qGapb +tGapb) == 0:
			#what to do then ...?, don't think, this happpens
			align_info.nr_paths_found -= 1
			continue

		(qStart, qEnd, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds, tExonSizes, tStarts, tEnds) =\
				 pu.get_splice_info(splice_align[i], est_align[i])
		#print tStarts
		
		score = alignment_scores[i]	
		
		(exon_length, identity, ssQuality, exonQuality, comparison, qIDX, tIDX) =\
					  pu.comp_identity(dna_letters[i], est_letters[i], t_strand, qStart, tStart, translation, comparison_char)

		first_identity = None
		last_identity = None 
		for i in range(len(comparison)):
		  if comparison[i] == '|' and first_identity is None: first_identity = i 
		  #print i,comparison[i],first_identity,qIDX[i],tIDX[i]
		  if comparison[-i] == '|' and last_identity is None: last_identity = len(comparison) - i - 1
		#print first_identity, last_identity, len(tIDX), len(qIDX)  
		#print tIDX[first_identity], tIDX[last_identity], qIDX[first_identity], qIDX[last_identity]
		#print tStart, tEnd, qStart, qEnd
		#exon_length[0] = exon_length[0] - tIDX[first_identity] - tStart + 2
		#exon_length[-1] = exon_length[-1] - tEnd + tIDX[last_identity] + 1
		tStart = tIDX[first_identity]
		tStarts[0] = tIDX[first_identity]
		tEnd = tIDX[last_identity]
		tEnds[-1] = tIDX[last_identity]
		qStart = qIDX[first_identity]
		qStarts[0] = qIDX[first_identity]
		qEnd = qIDX[last_identity]
		qEnds[-1] = qIDX[last_identity]

		align_info.score = score 
		align_info.match = match
		align_info.mism = mism
		align_info.qGapb = qGapb
		align_info.tGapb = tGapb
		align_info.qStart = qStart
		align_info.qEnd = qEnd
		align_info.tStart = tStart
		align_info.tEnd = tEnd
		align_info.num_exons = num_exons
		align_info.qExonSizes = qExonSizes
		align_info.qStarts = qStarts
		align_info.qEnds = qEnds
		align_info.tExonSizes =tExonSizes
		align_info.tStarts = tStarts
		align_info.tEnds = tEnds
		align_info.exon_length = exon_length
		align_info.identity = identity
		align_info.ssQuality = ssQuality
		align_info.exonQuality = exonQuality
		align_info.comparison = comparison
		align_info.qIDX = qIDX
		align_info.tIDX = tIDX 
		align_info.dna_letters = dna_letters
		align_info.est_letters = est_letters

		ai.append(align_info)

#<------------------------------------------------------------------------------>

	return ai
