# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# Written (W) 2006-2007 Soeren Sonnenburg
# Written (W) 2007 Gunnar Raetsch
# Copyright (C) 2007 Fraunhofer Institute FIRST and Max-Planck-Society
# 

import sys
import numpy
from numpy import mat,array,chararray,inf

class model(object):
	#acceptor
	acceptor_bins=None
	acceptor_limits=None
	acceptor_penalties=None

	acceptor_splice_b=None
	acceptor_splice_order=None
	acceptor_splice_window_left=None
	acceptor_splice_window_right=None
	acceptor_splice_alphas=None
	acceptor_splice_svs=None

	#donor
	donor_bins=None
	donor_limits=None
	donor_penalties=None
	
	donor_splice_b=None
	donor_splice_order=None
	#donor_splice_use_gc=None
	donor_splice_window_left=None
	donor_splice_window_right=None
	donor_splice_alphas=None
	donor_splice_svs=None

	# intron length
	intron_len_bins=None
	intron_len_limits=None
	intron_len_penalties=None
	intron_len_min=None
	intron_len_max=None
	intron_len_transform=None

	gene_len_max = None
	
	# substitution matrix
	substitution_matrix=None 

def parse_file(file):
	m=model()

	sys.stdout.write("loading model file"); sys.stdout.flush()

	l=file.readline();

	if l != '%palma definition file version: 1.0\n':
		sys.stderr.write("\nfile not a palma definition file\n")
		print l
		return None

	while l:
		if not ( l.startswith('%') or l.startswith('\n') ): # comment
			#print l[0:100]
			
			#acceptor
			if m.acceptor_bins is None: m.acceptor_bins=parse_value(l, 'acceptor_bins')
			if m.acceptor_limits is None: m.acceptor_limits=parse_vector(l, file, 'acceptor_limits')
			if m.acceptor_penalties is None: m.acceptor_penalties=parse_vector(l, file, 'acceptor_penalties') #DEAD

			if m.acceptor_splice_b is None: m.acceptor_splice_b=parse_value(l, 'acceptor_splice_b')
			if m.acceptor_splice_order is None: m.acceptor_splice_order=parse_value(l, 'acceptor_splice_order')
			if m.acceptor_splice_window_left is None: m.acceptor_splice_window_left=parse_value(l, 'acceptor_splice_window_left')
			if m.acceptor_splice_window_right is None: m.acceptor_splice_window_right=parse_value(l, 'acceptor_splice_window_right')
			if m.acceptor_splice_alphas is None: m.acceptor_splice_alphas=parse_vector(l, file, 'acceptor_splice_alphas')
			if m.acceptor_splice_svs is None: m.acceptor_splice_svs=parse_string(l, file, 'acceptor_splice_svs')

			#donor
			if m.donor_bins is None: m.donor_bins=parse_value(l, 'donor_bins')
			if m.donor_limits is None: m.donor_limits=parse_vector(l, file, 'donor_limits')
			if m.donor_penalties is None: m.donor_penalties=parse_vector(l, file, 'donor_penalties') #DEAD

			if m.donor_splice_b is None: m.donor_splice_b=parse_value(l, 'donor_splice_b')
			if m.donor_splice_order is None: m.donor_splice_order=parse_value(l, 'donor_splice_order')
			#if m.donor_splice_use_gc is None: m.donor_splice_use_gc=parse_value(l, 'donor_splice_use_gc')
			if m.donor_splice_window_left is None: m.donor_splice_window_left=parse_value(l, 'donor_splice_window_left')
			if m.donor_splice_window_right is None: m.donor_splice_window_right=parse_value(l, 'donor_splice_window_right')
			if m.donor_splice_alphas is None: m.donor_splice_alphas=parse_vector(l, file, 'donor_splice_alphas')
			if m.donor_splice_svs is None: m.donor_splice_svs=parse_string(l, file, 'donor_splice_svs')


			# intron length
			if m.intron_len_bins is None: m.intron_len_bins=parse_value(l, 'intron_len_bins')
			if m.intron_len_limits is None: m.intron_len_limits=parse_vector(l, file, 'intron_len_limits')
			if m.intron_len_penalties is None: m.intron_len_penalties=parse_vector(l, file, 'intron_len_penalties')
			if m.intron_len_min is None: m.intron_len_min=parse_value(l, 'intron_len_min')
			if m.intron_len_max is None: m.intron_len_max=parse_value(l, 'intron_len_max')
			if m.intron_len_transform is None: m.intron_len_transform=parse_value(l, 'intron_len_transform')

			# gene length
			if m.gene_len_max is None: m.gene_len_max=parse_value(l, 'gene_len_max')

			if m.substitution_matrix is None: m.substitution_matrix=parse_vector(l, file, 'substitution_matrix')

		l=file.readline()

	sys.stderr.write('done\n')
	return m

def parse_value(line, name):
	if (line.startswith(name)):
		#print 'found ' + name
		sys.stdout.write('.'); sys.stdout.flush()
		str = line[line.find('=')+1:-1] ;
		if str[0] == "'":
			return str[1:-1]
		else:
			#print float(str)
			return float(str)
	else:
		return None

def parse_vector(line, file, name):
    mat = parse_matrix(line, file, name)
    if mat is None:
     return mat
    else:
     mat = numpy.array(mat).flatten()
     return mat 

def parse_matrix(line, file, name):
	if (line.startswith(name)):
		sys.stdout.write('.'); sys.stdout.flush()
		if line.find(']') < 0:
			l=''
			while l is not None and l.find(']') < 0:
				line+=l
				l=file.readline()
			if l is not None and l.find(']') >= 0:
				line+=l

		if line.find(']') < 0:
			sys.stderr.write("matrix `" + name + "' ended without ']'\n")
			return None
		else:
			return mat(line[line.find('['):line.find(']')+1])
	else:
		return None

def parse_string(line, file, name):
    if (line.startswith(name)):
        sys.stdout.write('.'); sys.stdout.flush()
        l=''
        lines=[]
        while l is not None and l.find(']') < 0:
            if l:
                lines+=[list(l[:-1])]
            l=file.readline()

        if l.find(']') < 0:
            sys.stderr.write("string ended without ']'\n")
            return None
        else:
            seqlen=len(lines[0])
            num=len(lines)

            trdat = chararray((seqlen,num),1,order='FORTRAN')
            for i in xrange(num):
                trdat[:,i]=lines[i]
            return trdat
    else:
        return None

if __name__ == '__main__':
	import bz2
	import sys
	#import hotshot, hotshot.stats

	def load():
		#f=bz2.BZ2File('../../splicing/msplicer/standalone/python/data/msplicer_elegansWS120_gc=1_orf=0.dat.bz2');
		filename='test.dat.bz2' 
		if filename.endswith('.bz2'):
			f=bz2.BZ2File(filename);
		else:
			f=file(filename);
			
		m=parse_file(f);

		print m.acceptor_bins is None
		print m.acceptor_limits is None
		print m.acceptor_penalties is None

		print m.acceptor_splice_b is None
		print m.acceptor_splice_order is None
		print m.acceptor_splice_window_left is None
		print m.acceptor_splice_window_right is None
		print m.acceptor_splice_alphas is None
		print m.acceptor_splice_svs is None

		print m.donor_bins is None
		print m.donor_limits is None
		print m.donor_penalties is None

		print m.donor_splice_b is None
		print m.donor_splice_order is None
		#print m.donor_splice_use_gc is None
		print m.donor_splice_window_left is None
		print m.donor_splice_window_right is None
		print m.donor_splice_alphas is None
		print m.donor_splice_svs is None
		print m.intron_len_bins is None
		print m.intron_len_limits is None
		print m.intron_len_penalties is None
		print m.intron_len_min is None
		print m.intron_len_max is None
		print m.intron_len_transform is None

		print m.substitution_matrix is None

		print m.intron_len_transform
		print m.substitution_matrix

	load()

	#prof = hotshot.Profile("model.prof")
	#benchtime = prof.runcall(load)
	#prof.close()
	#stats = hotshot.stats.load("model.prof")
	#stats.strip_dirs()
	#stats.sort_stats('time', 'calls')
	#stats.print_stats(20)
