#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# Written (W) 2006-2007 Uta Schulze, Gunnar Raetsch
# Copyright (C) 2006-2007 Max-Planck-Society
# 

import pdb

def print_line(score, match, mism, qGapb, tGapb, strand, tStrand, qName, qSize, qStart, qEnd,\
			   tName, tSize, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds,\
			   tExonSizes, tStarts, tEnds, exon_length, identity, out_file):
    """Produces a tab separated line in the out_file which describes the alignment.
    
    1. align. score: palma alignment score
    2. match:        number of matches in the local alignment
    3. mismatch:     number of mismatches in the local alignment 
    4. Q gap bases:  number of gaps on the target sequence (insert in query acc. to BLAT) 
    5. T gap bases:  number of gaps on the query sequence (insert in target acc. to BLAT) 
                    (introns, which are basically gaps on the query, are not included!) 
    6. strand:       '+' or '-' for query strand (second strand for target if provided)
                    if you align two "short" sequences and no whole genome search, the
                    strand equals "+" by default and logic.
    7. Q name:       first word in header of the entry in the query-file.
    8. Q size:       length of the sequence of the entry in the query-file
    9. Q start:      alignment start position on query sequence
   10. Q end:        alignment end position on query sequence
   11. T name:       first word in header of the entry in the target-file.
   12. T size:       length of the sequence of the entry in the target-file
   13. T start:      alignment start position on target sequence
   14. T end:        alignment end position on target sequence
   15. exon count:   number of exons in the alignment. Different to BLAT, where each block is
                    defined by the surrounding gaps, exons are defined by adjacent introns.
                    Exons can contain gaps. The length of the without gaps on both target
                    and query will most likely be different.
   16. Q bound:      Exact beginning and end of each exon on the query, without showing the gaps.
                    E.g. "ex1_start-ex1_end,ex2_start-ex2_end,..,"
                    Beginning of first exon most likely equals Q start, except the alignment
                    begins with an intron. Analog with Q end for the end of the last exon.
   17. T bound:      Exact beginning and end of each exon on the query, without showing the gaps.
                    Beginning of first exon most likely equals T start, except the alignment
                    begins with an intron. Analog with T end for the end of the last exon.
   18. exon lengths: These are actually the length of the exons in the alignment, where gaps are
                    included.
   19. identity:     Number of matches divided by length of exon on alignment times 100.
                    If the exon is perfect, and there are no gaps or mismatches, it's 100%.
    """

    qExonSizes_string = ''
    for i in qExonSizes:
        qExonSizes_string = qExonSizes_string + ('%d,' %i)
        tExonSizes_string = ''
    for i in tExonSizes:
        tExonSizes_string = tExonSizes_string + ('%d,' %i)
        qBounds_string = ''
    for (i,j) in zip(qStarts, qEnds):
        qBounds_string = qBounds_string + ('%d-%d,' %(i,j))
        tBounds_string = ''
    for (i,j) in zip(tStarts, tEnds):
        tBounds_string = tBounds_string + ('%d-%d,' %(i,j))
        exon_length_string = ''	
    for i in exon_length:
        exon_length_string = exon_length_string + ('%d,' %i)
        identity_string = ''
    for i in identity:
        identity_string = identity_string + ('%1.1f,' %(i*100))

    if tStrand == '+':
        out_file.write("%4.3f\t%d\t%d\t%d\t%d\t%s\t%s\t%d\t%d\t%d\t%s\t%d\t%d\t%d\t%d\t%s\t%s\t%s\t%s\n"\
                                   %(score, match, mism, qGapb, tGapb, strand, qName, qSize, qStart,\
                                         qEnd, tName, tSize, tStart, tEnd, num_exons, qBounds_string,\
                                         tBounds_string, exon_length_string, identity_string))
    if tStrand == '-':
        out_file.write("%4.3f\t%d\t%d\t%d\t%d\t%s%s\t%s\t%d\t%d\t%d\t%s\t%d\t%d\t%d\t%d\t%s\t%s\t%s\t%s\n"\
               %(score, match, mism, qGapb, tGapb, strand, tStrand, qName, qSize, qStart,\
                                         qEnd, tName, tSize, tStart, tEnd, num_exons, qBounds_string,\
                                         tBounds_string, exon_length_string, identity_string))
    return
#<------------------------------------------------------------------------------>

#<------------------------------------------------------------------------------>
# prints a line in the output_file which describes the alignment, in blat psl format
#<------------------------------------------------------------------------------>
def print_line_blatformat(score, match, mism, qGapb, tGapb, qStrand, tStrand, qName, qSize, qStart, qEnd,\
                          tName, tSize, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds,\
                          tExonSizes, tStarts, tEnds, exon_length, identity, ssQuality, out_file):
    """
    matches int unsigned ,           # Number of bases that match (in BLAT this excludes repeats, not here)
    misMatches int unsigned ,        # Number of bases that don't match
    repMatches int unsigned ,        # Number of bases that match but are part of repeats (always zero here)
    nCount int unsigned ,            # Number of 'N' bases (always zero here)
    qNumInsert int unsigned ,        # Number of inserts in query (length of tStarts)
    qBaseInsert int unsigned ,       # Number of bases inserted in query
    tNumInsert int unsigned ,        # Number of inserts in target (length of qStarts)
    tBaseInsert int unsigned ,       # Number of bases inserted in target
    strand char(2) ,                 # + or - for query strand, optionally followed by + or - for target strand
    qName varchar(255) ,             # Query sequence name
    qSize int unsigned ,             # Query sequence size
    qStart int unsigned ,            # Alignment start position in query
    qEnd int unsigned ,              # Alignment end position in query
    tName varchar(255) ,             # Target sequence name
    tSize int unsigned ,             # Target sequence size
    tStart int unsigned ,            # Alignment start position in target
    tEnd int unsigned ,              # Alignment end position in target
    blockCount int unsigned ,        # Number of blocks in alignment. A block contains no gaps.
    blockSizes longblob ,            # Size of each block in a comma separated list
    qStarts longblob ,               # Start of each block in query in a comma separated list
    tStarts longblob ,               # Start of each block in target in a comma separated list
    """

    if tStrand == '-':
        exon_length.reverse()
        qEnds.reverse()
        qStarts = map(lambda x: qSize - x + 1, qEnds)
        tStarts.reverse()
        tEnds.reverse()
    #if strand == '-':
    #    qStarts.reverse()
    #    qEnds.reverse()
    if True:
        for i in range(len(qStarts)):
            if i>0: qStarts[i]=qStarts[i-1]+exon_length[i-1]

    repMatches = 0
    nCount = 0
    qStarts_string = ''
    for i in qStarts:
        assert(i>0)
        qStarts_string = qStarts_string + ('%d,' %(i-1))
    tStarts_string = ''
    #if len(tStarts)>0:
    #    tStarts_string = tStarts_string + ('%d,' %(tStarts[0]))
    for i in tStarts:
        assert(i>0)
        tStarts_string = tStarts_string + ('%d,' %(i-1))
    exon_length_string = ''	
    for i in exon_length:
        exon_length_string = exon_length_string + ('%d,' %i)
    ssQuality_string = ''	
    for i in ssQuality:
        ssQuality_string = ssQuality_string + ('%1.1f,' %(i*100))
    identity_string = '' 
    for i in identity:
        identity_string = identity_string + ('%1.1f,' %(i*100))

    out_file.write('%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t'\
                   %(match, mism, repMatches, nCount, len(tStarts), qGapb, len(qStarts), tGapb))
    #out_file.write('%c%c\t' %(tStrand,qStrand))
    out_file.write('%c\t' %(tStrand))
        
    out_file.write('%s\t%d\t%d\t%d\t%s\t%d\t%d\t%d\t%d\t%s\t%s\t%s\t%s\t%s\n'\
                   %(qName, qSize, qStart-1, qEnd, tName, tSize, tStart-1, tEnd, \
                   num_exons, exon_length_string, qStarts_string, tStarts_string, identity_string, ssQuality_string))
        
    return
#<------------------------------------------------------------------------------>


#<------------------------------------------------------------------------------>
# prints the alignment, should be clear (hehe)
# len(DNA) == len(EST)
#<------------------------------------------------------------------------------>
def print_align(DNA, EST, comparison, qIDX, tIDX, out_file, width=60):
    import numpy
    idx = []
    idx_old = []
    for i in range(1,len(DNA),width): #e.g. 1, 61, 121 ...
        if i == 1:
            idx = [i,numpy.minimum(i+width-1,len(DNA))]
            out_file.write("Q: %9i  %s %9i\n" %(qIDX[idx[0]-1], EST[idx[0]-1:idx[1]], qIDX[idx[1]-1]))
            out_file.write("              %s     \n" %(comparison[idx[0]-1:idx[1]]))
            out_file.write("T: %9i  %s %9i\n\n\n" %(tIDX[idx[0]-1], DNA[idx[0]-1:idx[1]], tIDX[idx[1]-1]))
        else:
            idx_old = idx
            idx = [i,numpy.minimum(i+width-1,len(DNA))]
            out_file.write("Q: %9i  %s %9i\n" %(min(qIDX[idx_old[1]-1]+1,qIDX[idx[1]-1]), EST[idx[0]-1:idx[1]], qIDX[idx[1]-1]))
            out_file.write("              %s     \n" %(comparison[idx[0]-1:idx[1]]))
            out_file.write("T: %9i  %s %9i\n\n\n" %(min(tIDX[idx_old[1]-1]+1,tIDX[idx[1]-1]), DNA[idx[0]-1:idx[1]], tIDX[idx[1]-1]))
    return


def print_header(out_file):
    """prints some lines in output_file that describe what each column means

    Can be turned off via --noHead.
    """
    out_file.write("align.\tmatch\tmis-\tQ gap\tT gap\tstrand\tQ\tQ\tQ\tQ\tT\tT\tT\tT\texon\tQ\tT\texon\tidentity\n")
    out_file.write("score\t\tmatch\tbases\tbases\t\tname\tsize\tstart\tend\tname\tsize\tstart\tend\tcount\tbound\tbound\tlengths\n")
    out_file.write("--------------------------------------------------------------------------------------------------------------------------------------------------------\n")
    return

def print_blat_header(out_file):
    """prints blat header lines in output_file that describe what each column means

    Copied from blat output. Can be turned off via --noHead.
    """
    out_file.write('psLayout version 3\n\n')
    out_file.write("match\tmis-\trep.\tN's\tQ gap\tQ gap\tT gap\tT gap\tstrand\tQ\t\tQ\tQ\tQ\tT\t\tT\tT\tT\tblock\tblockSizes\tqStarts\t tStarts\n")
    out_file.write('\tmatch\tmatch\t\tcount\tbases\tcount\tbases\t\tname\t\tsize\tstart\tend\tname\t\tsize\tstart\tend\tcount\n')
    out_file.write('---------------------------------------------------------------------------------------------------------------------------------------------------------------\n')

def print_bed_header(bed_file):
    """prints some lines in bed_file that describe what each column means

    Can be turned off via --noHead.
    """
    bed_file.write("chrom\tchrom\tchrom\tname\tscore\tstrand\tblock\tblock\tblock\n")
    bed_file.write("\tStart\tEnd\t\t\t\tCount\tSizes\tStart\n")
    bed_file.write("---------------------------------------------------------------------\n")
    return



#<------------------------------------------------------------------------------>
# prints a line in the output bed_file which describes the alignment
#<------------------------------------------------------------------------------>
def print_bed_line(tName, tStart, tEnd, qName, score, tStrand,\
		   num_exons, tStarts, tEnds, bed_file):

    tStarts_string = ''
    tExonSizes_string = ''
    for (i,j) in zip(tStarts, tEnds):
        tStarts_string = tStarts_string + ('%d,' %(i))
	tExonSizes_string = tExonSizes_string + ('%d,' %(j-i+1))    

    bed_file.write("%s\t%d\t%d\t%s\t%4.3f\t%s\t%d\t%s\t%s\n"\
		%(tName, tStart, tEnd, qName, score, tStrand,\
		  num_exons, tExonSizes_string, tStarts_string))
    return
#<------------------------------------------------------------------------------>


def print_results(align_info, out_file, bed_file, tName, tSize, qName, qSize, qStrand, tStrand, parameters):
    """printing results in output_file"""

    score = align_info.score
    match = align_info.match
    mism = align_info.mism
    qGapb = align_info.qGapb
    tGapb = align_info.tGapb

    tStart = align_info.tStart
    tEnd = align_info.tEnd
    tStarts = align_info.tStarts
    tEnds = align_info.tEnds
    tIDX = align_info.tIDX
    
    qStart = align_info.qStart
    qEnd = align_info.qEnd
    qStarts = align_info.qStarts
    qEnds = align_info.qEnds
    qIDX = align_info.qIDX
    #if tStrand=='+':
    #    tStart = align_info.tStart
    #    tEnd = align_info.tEnd
    #    tStarts = align_info.tStarts
    #    tEnds = align_info.tEnds
    #    tIDX = align_info.tIDX
    #else:
    #    tEnd = tSize - align_info.tStart + 1
    #    tStart = tSize - align_info.tEnd + 1
    #    tStarts = map(lambda x: tSize - x + 1, align_info.tStarts)
    #    tEnds = map(lambda x: tSize - x + 1, align_info.tEnds)
    #    tIDX = map(lambda x: tSize - x + 1, align_info.tIDX)

    #if qStrand=='+':
    #    qStart = align_info.qStart
    #    qEnd = align_info.qEnd
    #    qStarts = align_info.qStarts
    #    qEnds = align_info.qEnds
    #    qIDX = align_info.qIDX
    #else:
    #    qEnd = qSize - align_info.qStart + 1
    #    qStart = qSize - align_info.qEnd + 1
    #    qStarts = map(lambda x: qSize - x + 1, align_info.qStarts)
    #    qEnds = map(lambda x: qSize - x + 1, align_info.qEnds)
    #    qIDX = map(lambda x: qSize - x + 1, align_info.qIDX)

    num_exons = align_info.num_exons
    qExonSizes = align_info.qExonSizes
    tExonSizes = align_info.tExonSizes
    exon_length = align_info.exon_length
    identity = align_info.identity
    ssQuality = align_info.ssQuality 
    exonQuality = align_info.exonQuality 
    comparison = align_info.comparison
    dna_letters = align_info.dna_letters
    est_letters = align_info.est_letters

    if parameters['print_psl2']:
        print_line(score, match, mism, qGapb, tGapb, qStrand, tStrand, qName, qSize, qStart, qEnd,\
                   tName, tSize, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds,\
                   tExonSizes, tStarts, tEnds, exon_length, identity, ssQuality, out_file)
    else:
        print_line_blatformat(score, match, mism, qGapb, tGapb, qStrand, tStrand, qName, qSize, qStart, qEnd,\
                   tName, tSize, tStart, tEnd, num_exons, qExonSizes, qStarts, qEnds,\
                   tExonSizes, tStarts, tEnds, exon_length, exonQuality, ssQuality, out_file)

    if parameters['print_sequences']:
        out_file.write("\n")
        print_align(dna_letters[0], est_letters[0], comparison, qIDX, tIDX, out_file)

    if print_header and bed_file is not None:
        print_bed_header(bed_file)
        print_bed_line(tName, tStart, tEnd, qName, score, tStrand, num_exons, tStarts, tEnds, bed_file)

