#!/usr/bin/env python2.4

import sys
#import os
import shutil
from distutils.core import setup, Extension

try:
    import numpy
except:
    sys.stderr.write("WARNING: did not find 'numpy'\n")

try:
    import shogun
except:
    sys.stderr.write("WARNING: did not find 'shogun'\n")
    sys.stderr.write("         shogun is required for computing splice site scores\n")
    sys.stderr.write("         See shogun website: http://www.shogun-toolbox.org)\n")

if sys.argv[1] != 'sdist':
    sys.stderr.write('Using swig to construct interface file alignment_tool.py\n')
    setup (name = 'palma._alignment_tool',
           version = '0.3.6',
           description = '!!!This is a hack to ensure that swig runs first. Real setup is below!!!',
           author = ['Uta Schulze', 'Bettina Hepp', 'Gunnar Raetsch', \
                     'Soeren Sonnenburg','Cheng Soon Ong'],
           author_email = ['Gunnar.Raetsch@tuebingen.mpg.de'],
           license='GPL2',
           url = 'http://www.fml.tuebingen.mpg.de/raetsch/projects/palma',
           ext_modules = [Extension('palma._alignment_tool',
                                    sources = ['src/bound_matrix.cpp', 'src/myalign_1best.cpp', \
                                               'src/fill_matrix_1best.cpp', 'src/result_align_1best.cpp',\
                                               'src/alignment_tool_1best.i'])],
           options={'build_ext':{'swig_opts':'-c++'}},
           )
    shutil.copy2('src/alignment_tool.py','palma/alignment_tool.py')
    #os.rename('src/alignment_tool.py','palma/alignment_tool.py')

sys.stderr.write('PALMA - mRNA to Genome Alignments using Large Margin Algorithms\n')
setup (name = 'palma',
       version = '0.3.6',
       description = 'PALMA - mRNA to Genome Alignments using Large Margin Algorithms',
       author = ['Uta Schulze', 'Bettina Hepp', 'Gunnar Raetsch', \
                 'Soeren Sonnenburg','Cheng Soon Ong'],
       author_email = ['Gunnar.Raetsch@tuebingen.mpg.de'],
       license='GPL2',
       url = 'http://www.fml.tuebingen.mpg.de/raetsch/projects/palma',
       #py_modules=['palma.palma_utils','palma.alignment_tool','palma.genomic',\
       py_modules=['palma.palma_utils','palma.genomic',\
                   'palma.model','palma.output_formating','palma.seqdict','palma.signal_detectors'],
       ext_modules = [Extension('palma._alignment_tool',
                                sources = ['src/bound_matrix.cpp', 'src/myalign_1best.cpp', \
                                           'src/fill_matrix_1best.cpp', 'src/result_align_1best.cpp',\
                                           'src/alignment_tool_1best.i'])],
       packages=['palma'],
       package_data={'palma': ['data/*.param.bz2']},
       scripts=['scripts/palma-align.py'],
       options={'build_ext':{'swig_opts':'-c++'}},
       long_description='''
       PALMA aligns two genetic sequences 'the best way' according
       to its underlying algorithm and trained parameters. See the paper on the website for details.''')

# cleanup swig outputs so that each build is clean
#os.remove('palma/alignment_tool.py')
#os.remove('src/alignment_tool_1best_wrap.cpp')
