/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2004-2006 Uta Schulze
 * Written (W) 2004-2006 Bettina Hepp
 * Written (W) 2004-2006 Gunnar Raetsch
 * Copyright (C) 2004-2006 Fraunhofer Institute FIRST and Max-Planck-Society
 */

/*********************************************************************************************/
// Version of fill_matrix with fills several alignment matrices (best, second_best ...)
// nr_paths det. number of these matrices
//
// est_len: Number of columns in alignment matrix: length(EST) +1 (for the gap) (variable: i)
// dna_len: Number of lines in alignment matrix: length(DNA) +1 (for the gap) (variable: j)
//
// !!! IMPORTANT !!! The pictures and equations in the paper/poster etc. describe the algorithm
// for an alignment matrix where DNA and EST changed places !!! IMPORTANT !!!
/*********************************************************************************************/

/*
  matchmatrix: columnwise
    - A C G T N (dna)
  - x x x x x x
  A x x x x x x	
  C x x x x x x
  G x x x x x x
  T x x x x x x
  N x x x x x x
(est)
*/

/*
  alignment matrix: columnwise
   |-EST . . .
  -+------------
  -|00 ...
  D|0.
  N|. .
  A|.  .
  .|.
  .|
  .|
*/


/* TODO
   - max_len = 10000 for elegans, means, score for intron of length > max_len is -inf: TODO
   - do not use double

*/


#include <math.h>
#include "myalign_nbest.h"
#include "Mathmatics.h"
#include <assert.h>
#include <algorithm> //Eigentlich in Header

using namespace std;

void fill_matrix(int nr_paths, Pre_score* matrices[], int est_len /*counter var: i*/, int dna_len /*counter var: j*/, char* est, char* dna, struct penalty h, double* matchmatrix, double* donor, double* acceptor, int nr_donor_sites, int* donor_sites, int* max_score_positions)
{	

  /*********************************************************************************************/
  // variables
  /*********************************************************************************************/
  const int mlen = 6; // length of matchmatrix
  int estChar, dnaChar; // 0:'-', 1:'a', 2:'c', 3:'g', 4:'t', 5:'n'
  double prevValue ;
  double tempValue;
  int last_supp_point = (int)h.limits[h.len-1] ; //last (len) supporting point of h
  //fprintf(stdout,"last supp_point in h: %d\n", last_supp_point) ;

  int num_elem ; // counter variable

  const int max_num_elem = (4+nr_paths+last_supp_point) * nr_paths ; // upper bound on num_elem
  double* scores = new double[max_num_elem] ; //are sorted later
  if (!scores) {throw MyAlignException() ;}
  struct ArrayElem *array=new ArrayElem[max_num_elem] ;
  if (!array) {throw MyAlignException() ;}
  //fprintf(stdout, "%d\n", max_num_elem*sizeof(ArrayElem)) ;
  
  int max_ss_element[nr_paths*nr_paths]; //position on dna with max ss score for introns of length greater than last_supp_point
  int splice_pos ; //splice position get from donor_sites
  int start_splice_pos ; //where to start looking in donor_sites

  double max_scores[nr_paths] ;

  /*********************************************************************************************/
  // Initialisation of all alignment matrices, columnwise
  /*********************************************************************************************/
  for (int z=0; z<nr_paths; z++) /* matrice z */
  {
    max_scores[z] = log(0.0) ;
    max_score_positions[2*z]    = 0; // pos i
    max_score_positions[2*z +1] = 0; // pos j

    for (int i=0; i<est_len; i++) /* i.column */
    {
      for (int j=0; j<dna_len; j++) /* j.line */
      {
	((Pre_score*)matrices[z] +(i*dna_len) +j)->value = log(0.0); // -inf
	((Pre_score*)matrices[z] +(i*dna_len) +j)->prev_i = 0;
	((Pre_score*)matrices[z] +(i*dna_len) +j)->prev_j = 0;
	((Pre_score*)matrices[z] +(i*dna_len) +j)->prev_matrix_no = 0;
	((Pre_score*)matrices[z] +(i*dna_len) +j)->issplice = 0;
      }
    }
  }

	
  /*********************************************************************************************/	
  // Element (0,0) for first/best matrice (z=0)
  /*********************************************************************************************/	 
  ((Pre_score*)matrices[0])->value = 0;
  ((Pre_score*)matrices[0])->prev_i = 0;
  ((Pre_score*)matrices[0])->prev_j = 0;
  ((Pre_score*)matrices[0])->prev_matrix_no = 0;
  ((Pre_score*)matrices[0])->issplice = 0;
 

  /*********************************************************************************************/	
  // Line Zero for first/best matrice (z=0)
  /*********************************************************************************************/
  for (int i=1; i<est_len; i++) /* 0.line, all columns (DNA='-') */
  { 	
    ((Pre_score*)matrices[0] +(i*dna_len))->value = 0 ;		
    ((Pre_score*)matrices[0] +(i*dna_len))->prev_i = i-1;
    ((Pre_score*)matrices[0] +(i*dna_len))->prev_j = 0;    
    ((Pre_score*)matrices[0] +(i*dna_len))->prev_matrix_no = 0;
    ((Pre_score*)matrices[0] +(i*dna_len))->issplice = 0;
  }


  /*********************************************************************************************/	
  // Column Zero for first/best matrice (z=0)
  /*********************************************************************************************/	
  for (int j=1; j<dna_len; j++) // 0.column, all lines
  {      
    ((Pre_score*)matrices[0] +j)->value =  0 ;
    ((Pre_score*)matrices[0] +j)->prev_i = 0 ;
    ((Pre_score*)matrices[0] +j)->prev_j = j-1 ;
    ((Pre_score*)matrices[0] +j)->prev_matrix_no = 0 ;
    ((Pre_score*)matrices[0] +j)->issplice = 0;
  }

   
  /*********************************************************************************************/	
  // !!!! Column Zero and line Zero for all other matrices except the best one (z>0): -INF !!!! 
  /*********************************************************************************************/	


  
  /*********************************************************************************************/
  // all other entries for all matrices, last column and row treated seperately
  /*********************************************************************************************/
  
  /* columnwise, 0.column and 0.line already filled */
  for (int i=1; i<est_len; i++) /* i.column */
  {
    
    start_splice_pos = 0 ; //where to start looking in donor_sites
    for (int zz=0; zz<nr_paths*nr_paths; zz++) {max_ss_element[zz]=0 ;} // initializing
  
    for (int j=1; j<dna_len; j++) //j.line
    {
      dnaChar = check_char(dna[j-1]) ; // -1 because of of the gaps in the 0.line/column of the al-matrix
      estChar = check_char(est[i-1]) ;

      
      //if introns of length greater than last_supp_point possible
      if (j > last_supp_point)
      {
	//storing highest ss_score (the position) and NOT -INF
	int new_ss_pos = j-last_supp_point ;
	double donor_score = donor[new_ss_pos-1] ;
	if (isfinite(donor_score)) //splice site!
        {
	  if (max_ss_element[nr_paths-1] != 0)
          { 
	    for (int zz=0; zz<nr_paths; zz++) //for each alignment matrix
            {
	      int pos_with_minscore = nr_paths ; //assuming, the new score is the worst score
	      double new_intron_score = donor_score+((Pre_score*)matrices[zz]+(i*dna_len)+new_ss_pos-1)->value ; //could be -INF
	      if (!(isfinite(new_intron_score))) {fprintf(stdout, "new_intron_score is -INF\n") ;}
	      for (int pos=0; pos<nr_paths; pos++) //find worst score of all nr_paths+1 scores
              {
		int act_splice_position = max_ss_element[zz*nr_paths+pos] ;
		assert(isfinite(donor[act_splice_position-1])) ;
		double act_intron_score = donor[act_splice_position-1]+((Pre_score*)matrices[zz]+(i*dna_len)+act_splice_position-1)->value ;
		if (new_intron_score >= act_intron_score) 
                {
		  new_intron_score = act_intron_score ;
		  pos_with_minscore = pos ;
		} //else nothing happens, new intron score still the worst one
	      }
	      if (pos_with_minscore != nr_paths)
              {
		max_ss_element[zz*nr_paths+pos_with_minscore] = new_ss_pos ; //replacing
	      }
	    }
	  }
	  else
          {
	    for (int pos=0; pos<nr_paths; pos++)
            {
	      if (max_ss_element[pos]==0)  //no previous splice site at this pos
              {
		for (int zz=0; zz<nr_paths; zz++) {max_ss_element[zz*nr_paths+pos]=j-last_supp_point ;}
		break ;
	      }
	    } 
	  }
	}
	//start_pos in donor_sites so that only introns shorter then "last_supp_point" are seen
	if ((start_splice_pos < nr_donor_sites) && donor_sites[start_splice_pos] == j-last_supp_point)
        {
	  start_splice_pos++ ;
	}
      }
      

      /* Computing all possible scores */
      num_elem = 0 ; /* number of different scores/possibilities so far */
      for (int z=0; z<nr_paths; z++) /* matrice z */
      {
	Pre_score* actMatrix = (Pre_score*)matrices[z]; /* hm, why?  */

	/***************************************/
	// 0. Zero (local alignment version)
        /***************************************/
	assert(num_elem<max_num_elem) ;
	array[num_elem].prev_i = 0; 
	array[num_elem].prev_j = 0; 
	array[num_elem].prev_matrix_no = z;
	array[num_elem].isSplice = false; /* no splice site */
	scores[num_elem] = 0 ;
	num_elem++ ;


        /***************************************/
	// 1. gap on EST (j-1,i) -> (j,i) 
        /***************************************/
	prevValue = ((Pre_score*)actMatrix +(i*dna_len)+j-1)->value ; /* (j-1,i) -> (j,i) */
	if (isfinite(prevValue))
	{
	  tempValue = prevValue +(matchmatrix[mlen* dnaChar]); /* score(gap,DNA) */
	  if (isfinite(tempValue))
	  {
	    assert(num_elem<max_num_elem) ;
	    array[num_elem].prev_i = i; 
	    array[num_elem].prev_j = j-1; /* predecessor */
	    array[num_elem].prev_matrix_no = z;
	    array[num_elem].isSplice = false; /* no splice site */
	    scores[num_elem] = -tempValue ;
	    num_elem++ ;
	  }
	}
      

	/***************************************/
    	// 2. match/mismatch (j-1,i-1) -> (j,i)
	/***************************************/
	prevValue = ((Pre_score*)actMatrix +(i-1)*dna_len + j-1)->value ; /* (j-1,i-1) -> (j,i) */
	if (isfinite(prevValue))
	{
	  tempValue = prevValue +(matchmatrix[mlen* dnaChar +estChar]);//DNA mit EST
	  if (isfinite(tempValue))
	  {
	    assert(num_elem<max_num_elem) ;
	    array[num_elem].prev_i = i-1; /* predecessor */
	    array[num_elem].prev_j = j-1; /* predecessor */
	    array[num_elem].prev_matrix_no = z;
	    array[num_elem].isSplice = false; /* no splice site */
	    scores[num_elem] = -tempValue ;
	    num_elem++ ;
	  }
	}
	

	/***************************************/
	// 3. gap on DNA (j,i-1) -> (j,i)
	/***************************************/
	prevValue = ((Pre_score*)actMatrix +(i-1)*dna_len + j)->value ;
	if (isfinite(prevValue))
	{
	  tempValue = prevValue +(matchmatrix[estChar]); /* score(EST,gap) */
	  if (isfinite(tempValue))
	  {
	    assert(num_elem<max_num_elem) ;
	    array[num_elem].prev_i = i-1; /* predecessor */
	    array[num_elem].prev_j = j; 
	    array[num_elem].prev_matrix_no = z;
	    array[num_elem].isSplice = false;  /* no splice site */
	    scores[num_elem]=-tempValue ;
	    num_elem++ ;
	  }
	  
	}


	/***************************************/
	// 4. all possible splice sites/introns
	/***************************************/
	// acceptor[j-1]: g of ag (-1 because cpp starts counting at 0)
	// donor[splice_pos-1]: g of gt/gc (-1 because cpp starts counting at 0)
	// donor_sites: the g's of the gt/gc
	
	if (isfinite(acceptor[j-1]))
	{
	  
	  //the nr_paths many introns of length greater than last_supp_point
	  for (int pos=0; pos<nr_paths; pos++)
          {
	    if (max_ss_element[z*nr_paths+pos]>0)
            {
	      assert((j-max_ss_element[z*nr_paths+pos]+1) > last_supp_point) ;
	      splice_pos = max_ss_element[z*nr_paths+pos] ; //the g of the gt 
	      prevValue = ((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->value ; //one before the _gt
	      if (isfinite(prevValue) && (((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->issplice!=1))
              {
		int dist = j -(splice_pos-1); //_gt...ag: dist from g to g
		assert(dist>last_supp_point) ;
		tempValue = penalty_lookup(h,dist) +donor[splice_pos-1] +acceptor[j-1] +prevValue;
		if (isfinite(tempValue))
                {
		  assert(num_elem<max_num_elem) ;
		  array[num_elem].prev_i = i; 
		  array[num_elem].prev_j = splice_pos-1; // predecessor
		  array[num_elem].prev_matrix_no = z;
		  array[num_elem].isSplice = true; // splice site
		  scores[num_elem]=-tempValue ;
		  num_elem++ ;
		}
		else if (isfinite(penalty_lookup(h,dist)))
                {
		  fprintf(stdout, "Something wrong with donor_sites(1000er) (intron_length: %d)\n", dist) ;
		  sleep(5000) ;
		}
	      }
	    }
	  }

	  

          //all introns of length <= last_supp_point
          for (int k=start_splice_pos; k < nr_donor_sites; k++) 
	  //for (int k=0; k < nr_donor_sites; k++) //also intron of length greater than 1000
	  {   
	    splice_pos = donor_sites[k] ;
	    if (splice_pos > j-4) { break ;}
	    prevValue = ((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->value ; //one before the _gt
	    
	    if (isfinite(prevValue) && (((Pre_score*)actMatrix + (i*dna_len) +splice_pos-1)->issplice!=1))
	    {
	      int dist = j -(splice_pos-1); //_gt...ag: dist from g to g
	      assert(dist<=1000) ;
	      tempValue = penalty_lookup(h,dist) +donor[splice_pos-1] +acceptor[j-1] +prevValue;
	      if (isfinite(tempValue))
	      {
		assert(num_elem<max_num_elem) ;
		array[num_elem].prev_i = i; 
		array[num_elem].prev_j = splice_pos-1; /* predecessor */
		array[num_elem].prev_matrix_no = z;
		array[num_elem].isSplice = true; /* splice site */
		scores[num_elem]=-tempValue ;
		num_elem++ ;
	      }
	      else if (isfinite(penalty_lookup(h,dist)))
	      {
		fprintf(stdout, "Something wrong with donor_sites (intron_length: %d)\n", dist) ;
		sleep(5000) ;
	      }
	    }
	  }
	}

      } //end of z
      
      CMath::nmin<struct ArrayElem>(scores, array, num_elem, nr_paths);
      
      /* the nr_paths many highest scores are written in the nr_paths many matrices*/
      for (int z=0; z<nr_paths; z++) {
	if (z>=num_elem) {
	  //nothing happens, entry in matrices[z] at pos. (i,j) still -inf
	}
	else {
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->value = -scores[z] ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->prev_i = array[z].prev_i ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->prev_j = array[z].prev_j ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->issplice = array[z].isSplice ;
	  ((Pre_score*)matrices[z] + (i*dna_len) +j)->prev_matrix_no = array[z].prev_matrix_no;

	  if (-scores[z]>=max_scores[z]) { //storing the postion of the highest alignment score yet
	    max_scores[z] = -scores[z] ;
	    max_score_positions[2*z]    = i; // pos i
	    max_score_positions[2*z +1] = j; // pos j
	  }
	}
      } //end of z

    } //end of j
  } // end of i



  //free memory
  delete[] array;
  delete[] scores;
  
  return;
}


