/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 1999-2006 Soeren Sonnenburg
 * Written (W) 1999-2006 Gunnar Raetsch
 * Copyright (C) 1999-2006 Fraunhofer Institute FIRST and Max-Planck-Society
 */

#include "io.h"
#include "common.h"

#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>

FILE* CIO::target=stdout;

CIO::CIO()
{
}

void CIO::message(EMessageType prio, const CHAR *fmt, ... )
{
	check_target();
	print_message_prio(prio, target);
    va_list list;
    va_start(list,fmt);
    vfprintf(target,fmt,list);
    va_end(list);
    fflush(target);
}

void CIO::buffered_message(EMessageType prio, const CHAR *fmt, ... )
{
	check_target();
	print_message_prio(prio, target);
    va_list list;
    va_start(list,fmt);
    vfprintf(target,fmt,list);
    va_end(list);
}

CHAR* CIO::skip_spaces(CHAR* str)
{
	INT i=0;

	if (str)
	{
		for (i=0; isspace(str[i]); i++);

		return &str[i];
	}
	else 
		return str;
}

void CIO::set_target(FILE* t)
{
	target=t;
}

void CIO::check_target()
{
	if (!target)
		target=stdout;
}

void CIO::print_message_prio(EMessageType prio, FILE* target)
{
	switch (prio)
	{
		case M_DEBUG:
			fprintf(target, "[DEBUG] ");
			break;
		case M_PROGRESS:
			fprintf(target, "[PROGRESS] ");
			break;
		case M_INFO:
			//fprintf(target, "[INFO] ");
			break;
		case M_NOTICE:
			fprintf(target, "[NOTICE] ");
			break;
		case M_WARN:
			fprintf(target, "[WARN] ");
			break;
		case M_ERROR:
			fprintf(target, "[ERROR] ");
			break;
		case M_CRITICAL:
			fprintf(target, "[CRITICAL] ");
			break;
		case M_ALERT:
			fprintf(target, "[ALERT] ");
			break;
		case M_EMERGENCY:
			fprintf(target, "[EMERGENCY] ");
			break;
		case M_MESSAGEONLY:
			break;
		default:
			break;
	}
}
