/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2004-2006 Uta Schulze
 * Written (W) 2004-2006 Bettina Hepp
 * Written (W) 2004-2006 Gunnar Raetsch
 * Copyright (C) 2004-2006 Fraunhofer Institute FIRST and Max-Planck-Society
 */

/***************************************************************************/
// Description 
/***************************************************************************/


#include <stdlib.h>
#include "myalign_1best.h"
#include <string.h>
#include <assert.h>

using namespace std ;


int check_char(char base)
{
  switch(base) {
  case 'A':
  case 'a': return 1;
  case 'C': 	
  case 'c': return 2;
  case 'G':
  case 'g': return 3;
  case 'T':
  case 't': return 4;
  case 'N':
  case 'n': return 5; 
  default : return 0;
  } 
}

double penalty_lookup(struct penalty h, double value)
{
  double ret ;

  if (value<h.min_len)
    ret = log(0.0) ;
  if (value>h.max_len)
    ret = log(0.0) ;

  if (strcmp(h.transform, "")==0)
    {}
  else if (strcmp(h.transform, "log(+1)")==0)
    value = log(value+1) ;
  else if (strcmp(h.transform, "log(+3)")==0)
    value = log(value+3) ;
  else if (strcmp(h.transform, "(+3)")==0)
    value = value+3 ;
  else if (strcmp(h.transform, "log")==0)
    value = log(value) ;
  else {
    fprintf(stderr,"unknown transform\n") ;
    //CIO::message(M_ERROR, "unknown transform\n") ;
  }

  int upper_idx = -1;
  int lower_idx = -1;

  if (value <= h.limits[0]) {
    ret = h.penalties[0] ;
  }
  else if (value >= h.limits[h.len-1]) {
    ret = h.penalties[h.len-1] ;
  }
  else {
    for (int i=0; i<h.len; i++) {
      if (value <= h.limits[i]) {
	upper_idx = i ;
	break ;
      }
    }
    lower_idx = upper_idx-1 ;
    ret = ((value-h.limits[lower_idx])*h.penalties[upper_idx] + (h.limits[upper_idx]-value)*h.penalties[lower_idx])/(h.limits[upper_idx]-h.limits[lower_idx]) ;
  }
  return ret ;
}


struct align_info myalign(int reduce_region, const int mlen, int dna_len, char* dna, int est_len, char* est, struct penalty h, double* matchmatrix, double* donor, double* acceptor)
{
	int donor_length = dna_len ;
	int nr_paths = 1 ;
	
	dna_len++ ;  //is length(dna)+ 1 for NULL at end of string
	est_len++ ;  //is length(est)+ 1 for NULL at end of string
	struct align_info AIfP ;
	
	AIfP.alignment_length = new int[nr_paths] ;
	AIfP.alignment_scores = new double[nr_paths] ;
	if (!AIfP.alignment_length) {throw MyAlignException() ;}
	struct dnaest dna_est[nr_paths] ;
	
	double al_score ;
	
	/* some information 
	   cout << h.len << endl ;
	   cout << h.max_len << endl ;
	   cout << h.min_len << endl ;
	   cout << h.transform << endl ;
	   for (int i=0; i<30; i++) {
	   fprintf(stdout, "entry %d: <%f>\n", i, h.limits[i]);
	   }
	   for (int i=0; i<30; i++) {
	   fprintf(stdout, "entry %d: <%f>\n", i, h.penalties[i]);
	   }
	   
	   for (int i=0; i<30; i++) {
	   fprintf(stdout, "entry %d: <%f>\n", i, donor[i]);
	   }
	   for (int i=0; i<30; i++) {
	   fprintf(stdout, "entry %d: <%f>\n", i, acceptor[i]);
	   }
	   for (int i=0; i<mlen*mlen; i++) {
	   fprintf(stdout, "entry %d: <%f>\n", i, matchmatrix[i]);
	   }
	   cout << "dna: " << dna << endl ;
	   cout << "est: " << est << endl ;
	*/

	int est_len_cut = est_len ;
	int dna_len_cut = dna_len ;
	char dna_cut[dna_len] ;
	strncpy(dna_cut, dna, dna_len) ;
	char est_cut[est_len] ;
	strncpy(est_cut, est, est_len) ;
	int max_score_position_alstart[2] ;
	max_score_position_alstart[0] = 0 ;
	max_score_position_alstart[1] = 0 ;
	int max_score_position_alend[2] ;
	max_score_position_alend[0] = est_len ;
	max_score_position_alend[1] = dna_len ;
	
	if (reduce_region)
	{
		fprintf(stdout, "reducing region\n") ;

		/***************************************************************************/
		// possible donor positions
		/***************************************************************************/
		int nr_donor_sites = 0 ;
		for (int ii=0; ii<donor_length; ii++) {
			if(isfinite(donor[ii])) {
				nr_donor_sites++ ;
			}
		}
		int donor_sites[nr_donor_sites] ;
		int donor_idx = 0 ;
		for (int ii=0; ii<donor_length; ii++) {
			if(isfinite(donor[ii])) {
				donor_sites[donor_idx] = ii+1 ;
				donor_idx++ ;
			}  
		}
		
		
		/***************************************************************************/
		// computes alignment end
		/***************************************************************************/
		
		//int max_score_position_alend[2]; 
		try{
			al_score = bound_matrix(est_len, dna_len, est, dna, h, matchmatrix, donor, acceptor, nr_donor_sites, donor_sites, max_score_position_alend) ;
		} catch(exception& e) {
			throw MyAlignException() ;
		}
		
		//just some info
		fprintf(stdout, "reduce: end of alignment (bound_matrix): [%d/%d, %d/%d]\n", max_score_position_alend[0], est_len, max_score_position_alend[1], dna_len) ;
		//fprintf(stdout, "al_score (bound_matrix): %f\n", al_score) ;

		/***************************************************************************/
		// computes alighment start
		/***************************************************************************/
		char* rev_dna = new char[dna_len-1] ; 
		for (int i=0; i<dna_len-1; i++) {
			rev_dna[i] = dna[dna_len-2-i] ; 
		}
		char* rev_est = new char[est_len-1] ; 
		for (int i=0; i<est_len-1; i++) {
			rev_est[i] = est[est_len-2-i] ; 
		}
		double* rev_donor = new double[donor_length] ; 
		for (int i=0; i<donor_length; i++) {
			rev_donor[i] = acceptor[donor_length-1-i] ; 
		}
		double* rev_acceptor = new double[donor_length] ; 
		for (int i=0; i<donor_length; i++) {
			rev_acceptor[i] = donor[donor_length-1-i] ; 
		}
		//possible rev_donor positions
		int nr_rev_donor_sites = 0 ;
		for (int ii=0; ii<donor_length; ii++) {
			if(isfinite(rev_donor[ii])) {
				nr_rev_donor_sites++ ;
			}
		}
		int rev_donor_sites[nr_rev_donor_sites] ;
		int rev_donor_idx = 0 ;
		for (int ii=0; ii<donor_length; ii++) {
			if(isfinite(rev_donor[ii])) {
				rev_donor_sites[rev_donor_idx] = ii+1 ;
				rev_donor_idx++ ;
			}  
		}
		
		//int max_score_position_alstart[2] ;
		try{
			al_score = bound_matrix(est_len, dna_len, rev_est, rev_dna, h, matchmatrix, rev_donor, rev_acceptor, nr_rev_donor_sites, rev_donor_sites, max_score_position_alstart) ;
		} catch(exception& e) {
			throw MyAlignException() ;
		}
		max_score_position_alstart[0] = (est_len-1) -max_score_position_alstart[0] +1 ;
		max_score_position_alstart[1] = (dna_len-1) -max_score_position_alstart[1] +1 ;
		
		delete[] rev_dna ;
		delete[] rev_est ;
		delete[] rev_donor ;
		delete[] rev_acceptor ;
		//just some info
		fprintf(stdout, "reduce: start of alignment (bound_matrix): [%d/%d, %d/%d]\n", max_score_position_alstart[0], est_len, max_score_position_alstart[1], dna_len) ;
		//fprintf(stdout, "al_score (bound_matrix): %f\n", al_score) ;
		
		
		/***************************************************************************/
		// cut dna and est and so on
		/***************************************************************************/
		
		dna_len_cut = max_score_position_alend[1] - max_score_position_alstart[1] +2 ; //1 longer than len(dna)
		if (dna_len_cut<=0) {throw MyAlignException() ;}
		strncpy(dna_cut, (dna +max_score_position_alstart[1] -1), dna_len_cut-1) ;
		
		est_len_cut = max_score_position_alend[0] - max_score_position_alstart[0] +2 ; //1 longer than len(est)
		if (est_len_cut<=0) {throw MyAlignException() ;}
		strncpy(est_cut, (est +max_score_position_alstart[0] -1), est_len_cut-1) ;
		
		donor = donor + max_score_position_alstart[1] -1 ;
		donor_length = dna_len_cut -1 ;
		acceptor = acceptor + max_score_position_alstart[1] -1 ;
		
	}
	

	/***************************************************************************/
	// possible donor positions
	/***************************************************************************/
	int nr_donor_sites = 0 ;
	for (int ii=0; ii<donor_length; ii++) {
		if(isfinite(donor[ii])) {
			nr_donor_sites++ ;
		}
	}
	int donor_sites[nr_donor_sites] ;
	int donor_idx = 0 ;
	for (int ii=0; ii<donor_length; ii++) {
		if(isfinite(donor[ii])) {
			donor_sites[donor_idx] = ii+1 ;
			donor_idx++ ;
		}  
	}
	
	Pre_score* matrix = new Pre_score[dna_len_cut * est_len_cut];
	int matrix_ld = dna_len_cut ;
	
	if (!matrix) {throw MyAlignException() ;}
	
	int max_score_position[2] ;
	try{
		al_score = fill_matrix(matrix, est_len_cut, dna_len_cut, est_cut, dna_cut, h, matchmatrix, donor, acceptor, nr_donor_sites, donor_sites, max_score_position) ;
	} catch(exception& e) {
		throw MyAlignException() ;
	}
	//fprintf(stdout, "al_score (fill_matrix): %f\n", al_score) ;

	int dna_bt_start, est_bt_start ;
	if (!reduce_region)
	{
		int i = max_score_position[0] ; //i (est)
		int j = max_score_position[1] ; //j (dna)
		int prev_i = ((Pre_score*)matrix + i*dna_len_cut + j)->prev_i; 
		int prev_j = ((Pre_score*)matrix + i*dna_len_cut + j)->prev_j; 
		int first_match_i = 0 ;
		int first_match_j = 0 ;
		
		while (1) {
			if ((prev_i == (i-1)) && (prev_j == (j-1))) { // match or mismatch
				first_match_i = i ;
				first_match_j = j ;
			}
			
			i = prev_i;
			j = prev_j;    
			prev_i = ((Pre_score*)matrix + i*dna_len_cut + j)->prev_i; 
			prev_j = ((Pre_score*)matrix + i*dna_len_cut + j)->prev_j; 

			if ((i==prev_i) && (j==prev_j))
				break ;
		}
		//just some info
		//fprintf(stdout, "start of alignment (fill_matrix): [%d, %d]\n", first_match_i, first_match_j) ;
		//fprintf(stdout, "end of alignment (fill_matrix): [%d/%d, %d/%d]\n", max_score_position[0], est_len_cut, max_score_position[1], dna_len_cut) ;

		max_score_position_alstart[0]  = first_match_i   ;
		max_score_position_alstart[1]  = first_match_j  ;
		max_score_position_alend[0] = max_score_position[0]  ;
		max_score_position_alend[1] = max_score_position[1]  ;

		est_bt_start = max_score_position_alend[0] + 1 ;
		dna_bt_start = max_score_position_alend[1] + 1 ;
		//fprintf(stderr, "est_bt_start=%i, dna_bt_start=%i\n", est_bt_start, dna_bt_start) ;

		if (1)
		{ // Well, I don't like this, but it was the quickest way to get the two versions producing the same output
			assert(dna_len_cut == dna_len) ;
			assert(est_len_cut == est_len) ;
			
			dna_len_cut = max_score_position_alend[1] - max_score_position_alstart[1] +2 ; //1 longer than len(dna)
			if (dna_len_cut<=0) {throw MyAlignException() ;}
			strncpy(dna_cut, (dna +max_score_position_alstart[1] -1), dna_len_cut-1) ;
			
			est_len_cut = max_score_position_alend[0] - max_score_position_alstart[0] +2 ; //1 longer than len(est)
			if (est_len_cut<=0) {throw MyAlignException() ;}
			strncpy(est_cut, (est +max_score_position_alstart[0] -1), est_len_cut-1) ;
			
			Pre_score* new_matrix = new Pre_score[dna_len_cut * est_len_cut];
			for (int i=0; i<est_len_cut; i++)
				for (int j=0; j<dna_len_cut; j++)
				{
					new_matrix[i*dna_len_cut + j].prev_i = matrix[(i+max_score_position_alstart[0] -1)*matrix_ld + (j+max_score_position_alstart[1] -1)].prev_i - max_score_position_alstart[0] + 1;
					new_matrix[i*dna_len_cut + j].prev_j = matrix[(i+max_score_position_alstart[0] -1)*matrix_ld + (j+max_score_position_alstart[1] -1)].prev_j - max_score_position_alstart[1] + 1;
				}
			
			delete[] matrix ;
			matrix = new_matrix ;
			matrix_ld = dna_len_cut ;
			
			max_score_position_alstart[0]  = first_match_i - max_score_position_alstart[0] +1 ;
			max_score_position_alstart[1]  = first_match_j - max_score_position_alstart[0] +1 ;
			max_score_position_alend[0] = max_score_position[0] - max_score_position_alstart[0] +1 ;
			max_score_position_alend[1] = max_score_position[1] - max_score_position_alstart[0] +1 ;

			dna_bt_start = dna_len_cut ;
			est_bt_start = est_len_cut ;
			//fprintf(stderr, "est_bt_start=%i, dna_bt_start=%i\n", est_bt_start, dna_bt_start) ;
		}
	}
	else
	{
		dna_bt_start = dna_len_cut ;
		est_bt_start = est_len_cut ;
	}
	
  
	/***************************************************************************/ 
	// return arguments etc.
	/***************************************************************************/
	
	int sum_of_result_lengths = 0; 
	double alignmentscore = al_score ;
	int startend[4] ;
	startend[0] = max_score_position_alstart[0] ;
	startend[1] = max_score_position_alstart[1] ;
	startend[2] = max_score_position_alend[0] ;
	startend[3] = max_score_position_alend[1] ;
	
	//SpliceAlign for DNA 
	int* splice_align = new int[dna_len] ;
	if (!splice_align) {throw MyAlignException() ;}
	//SpliceAlign for EST 
	int* est_align = new int[est_len];
	if (!est_align) { throw MyAlignException() ;}
	// how many matches, mismatches, gaps
	int* mmatrix_param = new int[mlen*mlen];
	if (!mmatrix_param) { throw MyAlignException() ;}
	
	memset((char*)splice_align, -1, dna_len*sizeof(int)); // fills splice_align with -1
	memset((char*)est_align, -1, est_len*sizeof(int)); // fills est_align with -1
	memset((char*)mmatrix_param, 0, mlen*mlen*sizeof(int)); //fills mmatrix_param with zeros
	
	int result_length = 0 ; //length of alignment (computed by backtracking)
	
	bool no_more_path = result_align(matrix, matrix_ld, est_len, dna_len, est_bt_start, dna_bt_start, &result_length, est_cut, dna_cut, splice_align, est_align, mmatrix_param, startend);      
	//no_more_path could be used; equals one if backtracking failed 
	//fprintf(stderr, "est_bt_start=%i, dna_bt_start=%i\n", est_bt_start, dna_bt_start) ;
	
	if (no_more_path) {}
	
	// dnaest
	int* DNA = new int[result_length] ;
	if (!DNA) { throw MyAlignException() ;}
	int* EST = new int[result_length] ;
	if (!EST) { throw MyAlignException() ;}
	
	//backtracking
	int i = est_bt_start -1 ; //i (est)
	int j = dna_bt_start -1 ; //j (dna)
	int prev_i = ((Pre_score*)matrix + i*matrix_ld +j)->prev_i; 
	int prev_j = ((Pre_score*)matrix + i*matrix_ld +j)->prev_j;
    
	//fprintf(stderr, "%i %i %i %i\n", i, j, prev_i, prev_j) ;
	for (int ii=result_length; ii>0; ii--) {
		if ((prev_i == (i-1)) && (prev_j == (j-1))) { // match or mismatch
			DNA[ii-1] = check_char(dna_cut[j-1]) ;
			EST[ii-1] = check_char(est_cut[i-1]) ;
		}
		else if ((prev_i == (i)) && (prev_j == (j-1))) {// gap on EST
			DNA[ii-1] = check_char(dna_cut[j-1]) ;
			EST[ii-1] = 0 ; //gap
		}
		else if ((prev_i == (i-1)) && (prev_j == (j))) { // gap on DNA
			DNA[ii-1] = 0 ; //gap
			EST[ii-1] = check_char(est_cut[i-1]) ;
		}
		else {// splice site
			for (; j > prev_j; j--) {
				DNA[ii-1] = check_char(dna_cut[j-1]) ;
				EST[ii-1] = 6 ; //intron
				ii-- ;
			}
			ii++ ; // last ii-- too much (because done in for-loop)	
		}
		
		i = prev_i;
		j = prev_j;    
		prev_i = ((Pre_score*)matrix + i*matrix_ld + j)->prev_i; 
		prev_j = ((Pre_score*)matrix + i*matrix_ld + j)->prev_j; 
		//fprintf(stderr, "%i %i %i %i %i\n", i, j, prev_i, prev_j, ii) ;
	}
	
	sum_of_result_lengths += result_length ;
	AIfP.alignment_length[0] = result_length ;
	dna_est[0].DNA = DNA ;
	dna_est[0].EST = EST ;
	
	//just some info
	/*
	  for (int z=0; z<nr_paths; z++) {
	  fprintf(stdout,"z=%d, DNA=[",z) ;
	  for (int i=0;i<AIfP.alignment_length[z];i++) {
      fprintf(stdout, "%d,", dna_est[z].DNA[i]) ; 
	  }
	  fprintf(stdout, "]\n") ;
	  }
	  for (int z=0; z<nr_paths; z++) {
	  fprintf(stdout,"z=%d, EST=[",z) ;
	  for (int i=0;i<AIfP.alignment_length[z];i++) {
      fprintf(stdout, "%d,", dna_est[z].EST[i]) ; 
	  }
	  fprintf(stdout, "]\n") ;
	  }
	*/
	
	AIfP.aligned_dna = new int[sum_of_result_lengths] ;
	if (!AIfP.aligned_dna) { throw MyAlignException() ;}
	
	AIfP.aligned_est = new int[sum_of_result_lengths] ;
	if (!AIfP.aligned_est) { throw MyAlignException() ;}
	
	int matrix_no = 0 ;
	int help_idx = 0 ;
	for (int i=0;i<sum_of_result_lengths;i++) {
		AIfP.aligned_dna[i] = dna_est[matrix_no].DNA[help_idx] ;
		AIfP.aligned_est[i] = dna_est[matrix_no].EST[help_idx] ;
		help_idx++ ;
		if (help_idx == AIfP.alignment_length[matrix_no]) {
			matrix_no++ ;
			help_idx = 0 ;
		}
	}
	
	//just some info
	/*
	  fprintf(stdout,"aligned_dna=[") ;
	  for (int i=0;i<sum_of_result_lengths;i++) {
	  fprintf(stdout, "%d,", AIfP.aligned_dna[i]) ; 
	  }
	  fprintf(stdout, "]\n") ;
	  fprintf(stdout,"aligned_est=[") ;
	  for (int i=0;i<sum_of_result_lengths;i++) {
	  fprintf(stdout, "%d,", AIfP.aligned_est[i]) ; 
	  }
	  fprintf(stdout, "]\n") ;
	*/
	
	
	AIfP.splice_align = splice_align ;
	AIfP.est_align = est_align ;
	AIfP.mmatrix_param = mmatrix_param ;
	AIfP.alignment_scores[0] = alignmentscore ;
	//fprintf(stdout, "alignment_score (final): %f\n", alignmentscore) ;
 
	//just some info
	/*
	  fprintf(stdout, "splice_align: [") ;
	  for (int i=0;i<dna_len-1;i++) {
	  fprintf(stdout, "%d,", AIfP.splice_align[i]) ;
	  }
	  fprintf(stdout, "]\n") ;
	*/

	//Free Memory
	for (int i=nr_paths-1;i>=0; i--)
	{
		delete[] matrix;
		delete[] dna_est[i].DNA ;
		delete[] dna_est[i].EST ;
	}
	

	return AIfP;    
	
}
