/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2004-2006 Uta Schulze
 * Written (W) 2004-2006 Bettina Hepp
 * Written (W) 2004-2006 Gunnar Raetsch
 * Copyright (C) 2004-2006 Fraunhofer Institute FIRST and Max-Planck-Society
 */

/***************************************************************************/
// Description 
/***************************************************************************/


#include <stdlib.h>
#include "myalign_nbest.h"

using namespace std ;


int check_char(char base)
{
  switch(base) {
  case 'A':
  case 'a': return 1;
  case 'C': 	
  case 'c': return 2;
  case 'G':
  case 'g': return 3;
  case 'T':
  case 't': return 4;
  case 'N':
  case 'n': return 5; 
  default : return 0;
  } 
}

double penalty_lookup(struct penalty h, double value)
{
  double ret ;

  if (value<h.min_len)
    ret = log(0.0) ;
  if (value>h.max_len)
    ret = log(0.0) ;

  if (strcmp(h.transform, "")==0)
    {}
  else if (strcmp(h.transform, "log(+1)")==0)
    value = log(value+1) ;
  else if (strcmp(h.transform, "log(+3)")==0)
    value = log(value+3) ;
  else if (strcmp(h.transform, "(+3)")==0)
    value = value+3 ;
  else if (strcmp(h.transform, "log")==0)
    value = log(value) ;
  else {
    fprintf(stderr,"unknown transform\n") ;
    //CIO::message(M_ERROR, "unknown transform\n") ;
  }

  int upper_idx = -1;
  int lower_idx = -1;

  if (value <= h.limits[0]) {
    ret = h.penalties[0] ;
  }
  else if (value >= h.limits[h.len-1]) {
    ret = h.penalties[h.len-1] ;
  }
  else {
    for (int i=0; i<h.len; i++) {
      if (value <= h.limits[i]) {
	upper_idx = i ;
	break ;
      }
    }
    lower_idx = upper_idx-1 ;
    ret = ((value-h.limits[lower_idx])*h.penalties[upper_idx] + (h.limits[upper_idx]-value)*h.penalties[lower_idx])/(h.limits[upper_idx]-h.limits[lower_idx]) ;
  }
  return ret ;
}


struct align_info myalign(int nr_paths, const int mlen, int dna_len, char* dna, int est_len, char* est, struct penalty h, double* matchmatrix, double* donor, double* acceptor)
{
  int donor_length = dna_len ;

  dna_len++ ;  //is length(dna)+ 1 for NULL at end of string
  est_len++ ;  //is length(est)+ 1 for NULL at end of string
  struct align_info AIfP ;

  AIfP.alignment_length = new int[nr_paths] ;
  if (!AIfP.alignment_length) {throw MyAlignException() ;}
  struct dnaest dna_est[nr_paths] ;

  /* some information
  cout << h.len << endl ;
  cout << h.max_len << endl ;
  cout << h.min_len << endl ;
  cout << h.transform << endl ;
  for (int i=0; i<30; i++) {
    fprintf(stdout, "entry %d: <%f>\n", i, h.limits[i]);
  }
  for (int i=0; i<30; i++) {
    fprintf(stdout, "entry %d: <%f>\n", i, h.penalties[i]);
  }

  for (int i=0; i<donor_length; i++) {
    fprintf(stdout, "entry %d: <%f>\n", i, donor[i]);
  }
  for (int i=0; i<acceptor_length; i++) {
    fprintf(stdout, "entry %d: <%f>\n", i, acceptor[i]);
  }
  for (int i=0; i<mlen*mlen; i++) {
    fprintf(stdout, "entry %d: <%f>\n", i, matchmatrix[i]);
  }
  cout << "dna: " << dna << endl ;
  cout << "est: " << est << endl ;
  */


 
  /***************************************************************************/ 
  // initialize alignment matrices and call fill_matrix()  
  /***************************************************************************/
  
  //possible donor positions
  int nr_donor_sites = 0 ;
  for (int ii=0; ii<donor_length; ii++) {
    if(isfinite(donor[ii])) {
      nr_donor_sites++ ;
    }
  }
  int donor_sites[nr_donor_sites] ;
  int donor_idx = 0 ;
  for (int ii=0; ii<donor_length; ii++) {
    if(isfinite(donor[ii])) {
      donor_sites[donor_idx] = ii+1 ;
      donor_idx++ ;
    }  
  }

  //local alignment: where is the highest alignment score [i1,j1,i2,j2...]
  int max_score_positions[nr_paths*2]; 

  Pre_score* matrices[nr_paths];
  try{
    for (int z=0; z<nr_paths; z++) {
      matrices[z] = new Pre_score[dna_len * est_len];
      if (!matrices[z]) { throw MyAlignException() ;}
      
      //fprintf(stdout,"size of one matrice: %d\n", est_len*dna_len*sizeof(Pre_score)) ;
    }
  } catch(exception& e) {
    throw MyAlignException() ;
  }
  

  try{
    fill_matrix(nr_paths, matrices, est_len, dna_len, est, dna, h, matchmatrix, donor, acceptor, nr_donor_sites, donor_sites, max_score_positions) ;
  } catch(exception& e) {
    throw MyAlignException() ;
  }

 
  /*
  //just some info
  for (int z=0; z<nr_paths; z++) {
    fprintf(stdout, "(%d, %d), est_len: %d, dna_len: %d\n", max_score_positions[2*z], max_score_positions[2*z +1], est_len-1, dna_len-1) ;
  }
  */
 
  /***************************************************************************/ 
  // return arguments etc.
  /***************************************************************************/

  int sum_of_result_lengths = 0; 
  
  //SpliceAlign for DNA - 1 line per matrix/path
  int* splice_align = new int[(dna_len-1)*nr_paths] ;
  if (!splice_align) {throw MyAlignException() ;}
  //SpliceAlign for EST - 1 line per matrix/path
  int* est_align = new int[(est_len-1)*nr_paths];
  if (!est_align) { throw MyAlignException() ;}
  // how many matches, mismatches, gaps - 1 row per matrix/path
  int* mmatrix_param = new int[(mlen*mlen)*nr_paths];
  if (!mmatrix_param) { throw MyAlignException() ;}
  ///alignment score for each path/matrix
  double* alignmentscores = new double[nr_paths];
  if (!alignmentscores) { throw MyAlignException() ;}
  
  memset((char*)splice_align, -1, (dna_len-1)*nr_paths*sizeof(int)); // fills splice_align with -1
  memset((char*)est_align, -1, (est_len-1)*nr_paths*sizeof(int)); // fills est_align with -1
  memset((char*)mmatrix_param, 0, mlen*mlen*nr_paths*sizeof(int)); //fills mmatrix_param with zeros
  memset(alignmentscores, -1, nr_paths*sizeof(double)); //fills alignmentscores with -1


 
  for (int z=0; z<nr_paths; z++) {	 
    int result_length = 0 ; //one variable for all matrices/paths
    
    int* s_align = splice_align + (dna_len-1)*z;  //pointer
    int* e_align = est_align + (est_len-1)*z ;    //pointer
    int* mparam  = mmatrix_param + (mlen*mlen)*z; //pointer
    
    bool no_more_path = result_align(matrices, z, est_len, dna_len, &result_length, est, dna, s_align, e_align, mparam, alignmentscores, max_score_positions);      
    //no_more_path could be used; equals one if backtracking failed 
    
    if (no_more_path) {}

    // dnaest
    int* DNA = new int[result_length] ;
    if (!DNA) { throw MyAlignException() ;}
    int* EST = new int[result_length] ;
    if (!EST) { throw MyAlignException() ;}

    //backtracking
    int i = max_score_positions[2*z] ; //i (est)
    int j = max_score_positions[2*z +1] ; //j (dna)
    int prev_i = ((Pre_score*)matrices[z] + i*dna_len +j)->prev_i; 
    int prev_j = ((Pre_score*)matrices[z] + i*dna_len +j)->prev_j;
    int prev_m_no = ((Pre_score*)matrices[z] + i*dna_len +j)->prev_matrix_no;
    
    for (int ii=result_length; ii>0; ii--) {
      if ((prev_i == (i-1)) && (prev_j == (j-1))) { // match or mismatch
	DNA[ii-1] = check_char(dna[j-1]) ;
	EST[ii-1] = check_char(est[i-1]) ;
      }
      else if ((prev_i == (i)) && (prev_j == (j-1))) {// gap on EST
	DNA[ii-1] = check_char(dna[j-1]) ;
	EST[ii-1] = 0 ; //gap
      }
      else if ((prev_i == (i-1)) && (prev_j == (j))) { // gap on DNA
	DNA[ii-1] = 0 ; //gap
	EST[ii-1] = check_char(est[i-1]) ;
      }
      else {// splice site
	for (; j > prev_j; j--) {
	  DNA[ii-1] = check_char(dna[j-1]) ;
	  EST[ii-1] = 6 ; //intron
	  ii-- ;
	}
	ii++ ; // last ii-- too much (because done in for-loop)	
      }
      
      i = prev_i;
      j = prev_j;    
      prev_i = ((Pre_score*)matrices[prev_m_no] + i*dna_len + j)->prev_i; 
      prev_j = ((Pre_score*)matrices[prev_m_no] + i*dna_len + j)->prev_j; 
      prev_m_no = ((Pre_score*)matrices[prev_m_no] + i*dna_len + j)->prev_matrix_no;
    }

    sum_of_result_lengths += result_length ;
    AIfP.alignment_length[z] = result_length ;
    dna_est[z].DNA = DNA ;
    dna_est[z].EST = EST ;
  } //end of z

  //just some info
  /*
  for (int z=0; z<nr_paths; z++) {
    fprintf(stdout,"z=%d, DNA=[",z) ;
    for (int i=0;i<AIfP.alignment_length[z];i++) {
      fprintf(stdout, "%d,", dna_est[z].DNA[i]) ; 
    }
    fprintf(stdout, "]\n") ;
  }
  for (int z=0; z<nr_paths; z++) {
    fprintf(stdout,"z=%d, EST=[",z) ;
    for (int i=0;i<AIfP.alignment_length[z];i++) {
      fprintf(stdout, "%d,", dna_est[z].EST[i]) ; 
    }
    fprintf(stdout, "]\n") ;
  }
  */
  

  AIfP.aligned_dna = new int[sum_of_result_lengths] ;
  if (!AIfP.aligned_dna) { throw MyAlignException() ;}

  AIfP.aligned_est = new int[sum_of_result_lengths] ;
  if (!AIfP.aligned_est) { throw MyAlignException() ;}

  int matrix_no = 0 ;
  int help_idx = 0 ;
  for (int i=0;i<sum_of_result_lengths;i++) {
    AIfP.aligned_dna[i] = dna_est[matrix_no].DNA[help_idx] ;
    AIfP.aligned_est[i] = dna_est[matrix_no].EST[help_idx] ;
    help_idx++ ;
    if (help_idx == AIfP.alignment_length[matrix_no]) {
      matrix_no++ ;
      help_idx = 0 ;
    }
  }

  //just some info
  /*
  fprintf(stdout,"aligned_dna=[") ;
  for (int i=0;i<sum_of_result_lengths;i++) {
    fprintf(stdout, "%d,", AIfP.aligned_dna[i]) ; 
  }
  fprintf(stdout, "]\n") ;
  fprintf(stdout,"aligned_est=[") ;
  for (int i=0;i<sum_of_result_lengths;i++) {
    fprintf(stdout, "%d,", AIfP.aligned_est[i]) ; 
  }
  fprintf(stdout, "]\n") ;
  */


  AIfP.splice_align = splice_align ;
  AIfP.est_align = est_align ;
  AIfP.mmatrix_param = mmatrix_param ;
  AIfP.alignment_scores = alignmentscores ;
  
  

  //just some info
  /*
  fprintf(stdout, "splice_align: [") ;
  for (int i=0;i<dna_len-1;i++) {
    fprintf(stdout, "%d,", AIfP.splice_align[i]) ;
  }
   fprintf(stdout, "]\n") ;
  */

  //Free Memory
  for (int i=nr_paths-1;i>=0; i--)
  {
    delete[] matrices[i];
    delete[] dna_est[i].DNA ;
    delete[] dna_est[i].EST ;
  }
  
  return AIfP;    
  
}
