/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2006      Uta Schulze
 * Written (W) 2004-2006 Bettina Hepp
 * Written (W) 2004-2006 Gunnar Raetsch
 * Copyright (C) 2004-2006 Fraunhofer Institute FIRST and Max-Planck-Society
 */

#include <assert.h>
#include "myalign_1best.h"

using namespace std;

bool result_align(Pre_score* matrix, int matrix_ld, int est_len, int dna_len, int est_len_cut, int dna_len_cut, int* result_length_ptr, char* est_cut, char* dna_cut, int* s_align, int* e_align, int* mparam, int* startend)
//Gibt 1 zurueck, wenn kein weiterer Pfad existiert, sonst 0
	
{  
	const int mlen=6; // length of matchmatrix
	// startend: [i_start, j_start, i_end, j_end] ;
	
	//fprintf(stdout, "in result_align: (%d, %d)\n",startend[2], startend[3]) ;
	int dnanum ; //using check_char: letter -> number
	int estnum ; //using check_char: letter -> number
	
	int dna_pos ;
	int est_pos ;
	int splice_state ;
	int est_state ;
	for (dna_pos=dna_len-1; dna_pos>startend[3]; dna_pos--) {
		s_align[dna_pos-1] = 4 ; // not aligned (-1 because array starts with index 0)
	}
	for (est_pos=est_len-1; est_pos>startend[2]; est_pos--) {
		e_align[est_pos-1] = 4 ;
	}
	//splice_align: 0: exon, 1: donor, 2: acceptor, 3: intro,. 4: dangling end
	//est_align: [1122221112...] for exons of length 2,4,3 ...
	
	/* starting point for backtracking */
	int i = est_len_cut -1 ; //length of matrix 
	int j = dna_len_cut -1 ; //length of matrix
	
	int prev_i = ((Pre_score*)matrix + i*matrix_ld +j)->prev_i; 
	int prev_j = ((Pre_score*)matrix + i*matrix_ld +j)->prev_j;
	
	
	splice_state = 0 ; //exon
	est_state = 1 ; //[112222111 ... for exons of length 2,4,3 ...
	
	//compute length of alignment (between max(m,n) and m+n)	
	//local_alignment: backtracking to first Zero: look at actual value
	while(!((i==0) && (j==0))) {
		
		//just some info
		//if ((z>0) && (!(z==prev_m_no))) {
		//  fprintf(stdout, "(%d,%d): next step in next matrix\n", i,j) ;
		//}
		
		if ((prev_i == (i-1)) && (prev_j == (j-1))) { // match or mismatch
			(*result_length_ptr)= (*result_length_ptr) + 1;
			
			s_align[dna_pos-1] = splice_state; 
			dna_pos-- ;
			e_align[est_pos-1] = est_state ; //1 or 2, depended
			est_pos-- ;
			
			dnanum = check_char(dna_cut[j-1]) ; //-1 because index starts with 0
			estnum = check_char(est_cut[i-1]) ; //-1 because index starts with 0
			mparam[mlen*dnanum +estnum] ++ ;
		}
		else if ((prev_i == (i)) && (prev_j == (j-1))) {// gap on EST
			(*result_length_ptr)= (*result_length_ptr) + 1;
			s_align[dna_pos-1] = splice_state; 
			dna_pos-- ;
			
			if (j>=startend[1])
			{
				dnanum = check_char(dna_cut[j-1]) ; //-1 because index starts with 0
				estnum = 0 ; //gap 
				mparam[mlen*dnanum +estnum] ++ ;
				//fprintf(stderr, "qgap %i %i \n", j, startend[1]) ;
			}
		}
		else if ((prev_i == (i-1)) && (prev_j == (j))) { // gap on DNA
			(*result_length_ptr)= (*result_length_ptr) + 1;
			
			e_align[est_pos-1] = est_state ; //1 or 2, depended
			est_pos-- ;
			
			dnanum = 0 ; //gap
			estnum = check_char(est_cut[i-1]) ; //-1 because index starts with 0
			mparam[mlen*dnanum +estnum] ++ ;
			//fprintf(stderr, "tgap\n", i, startend[0]) ;
		}
		else {// splice site
			if ((prev_i==0) && prev_j==0)
				break ;

			(*result_length_ptr) =  (*result_length_ptr) + (j - prev_j);
			/*if (((Pre_score*)matrix + i*dna_len_cut + j)->issplice==0) {
			  fprintf(stdout, "matrix(i,j): %3.15f\n",((Pre_score*)matrix + i*dna_len_cut +j)->value) ;
			  fprintf(stdout, "matrices[prev_m_no](i,j): %3.15f\n",((Pre_score*)matrices[prev_m_no] + i*dna_len_cut +j)->value) ;
			  fprintf(stdout, "issplice(act): %d\n",((Pre_score*)matrix + i*dna_len_cut + j)->issplice) ;
			  fprintf(stdout, "issplice(prev): %d\n",((Pre_score*)matrices[prev_m_no] + i*dna_len_cut + j)->issplice) ;
			  fprintf(stdout, "act_m_no: %d, prev_m_no: %d\n", act_m_no, prev_m_no) ;
			  fprintf(stdout, "j: %d -> prev_j:%d\n", j, prev_j) ;
			  fprintf(stdout, "i: %d -> prev_i:%d\n", i, prev_i) ;
			  }*/
			if (est_state==1) { //was exon labeled "1"
				est_state = 2 ; //new exon is labeled "2"
			}
			else {
				est_state = 1 ; //last exon was labeled "2", new exon is labeled "1"
			}
			for (; j > prev_j; j--) {
				if (splice_state == 0) { //coming from exon
					splice_state = 2 ; //first intron_state: acceptor
				} else {
					splice_state = 3 ; //intron
				}
				if (j-1 == prev_j) { //last intron_state: donor
					splice_state = 1 ;//donor
				}
				s_align[dna_pos-1] = splice_state; 
				dna_pos-- ;
			}
			splice_state = 0 ; //exon again
			
		}
		
		i = prev_i;
		j = prev_j;
	  
		prev_i    = ((Pre_score*)matrix + i*matrix_ld + j)->prev_i; 
		prev_j    = ((Pre_score*)matrix + i*matrix_ld + j)->prev_j; 

		//fprintf(stderr, "%i %i %i %i %i\n", i, j, prev_i, prev_j, (*result_length_ptr)) ;
		
	}
	for (; dna_pos>0; dna_pos--) {
		s_align[dna_pos-1] = 4 ; // not aligned (-1 because array starts with index 0)
	}
	for (; est_pos>0; est_pos--) {
		e_align[est_pos-1] = 4 ; // not aligned (-1 because array starts with index 0)
	}
	
	//fprintf(stdout, "(%d, %d) --> (%d, %d)\n", startend[0], startend[1], startend[2], startend[3]) ;
	//fprintf(stdout, "result_length: %d\n", *result_length_ptr) ;
	
	return 0;
}


