/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Written (W) 2004-2006 Uta Schulze
 * Written (W) 2004-2006 Bettina Hepp
 * Written (W) 2004-2006 Gunnar Raetsch
 * Copyright (C) 2004-2006 Fraunhofer Institute FIRST and Max-Planck-Society
 */

#include "assert.h"
#include "myalign_nbest.h"

using namespace std;

bool result_align(Pre_score* matrices[], int z, int est_len, int dna_len, int* result_length_ptr, char* est, char* dna, int* s_align, int* e_align, int* mparam, double* alignmentscores, int* max_score_positions)
//Gibt 1 zurueck, wenn kein weiterer Pfad existiert, sonst 0

{  
  const int mlen=6; // length of matchmatrix
  int startend[4] ; // [i_start, j_start, i_end, j_end] ;
  startend[3] = max_score_positions[2*z] ; //i (est)
  startend[4] = max_score_positions[2*z +1] ; //j (dna)
  //fprintf(stdout, "in result_align: z=%d:(%d, %d)\n", z, startend[3], startend[4]) ;
  int dnanum ; //using check_char: letter -> number
  int estnum ; //using check_char: letter -> number
  
  int dna_pos ;
  int est_pos ;
  int splice_state ;
  int est_state ;
  for (dna_pos=dna_len-1; dna_pos>startend[4]; dna_pos--) {
    s_align[dna_pos-1] = 4 ; // not aligned (-1 because array starts with index 0)
  }
  for (est_pos=est_len-1; est_pos>startend[3]; est_pos--) {
    e_align[est_pos-1] = 4 ;
  }
  //splice_align: 0: exon, 1: donor, 2: acceptor, 3: intro,. 4: dangling end
  //est_align: [1122221112...] for exons of length 2,4,3 ...
 
  /* starting point for backtracking */
  int i = startend[3] ; 
  int j = startend[4] ;
  int act_m_no  = z ;
  alignmentscores[z] = ((Pre_score*)matrices[z] + i*dna_len +j)->value;
  //fprintf(stdout, "alignment_score for z=%d: %3.8f\n", z, alignmentscores[z]) ;

  int prev_i    = ((Pre_score*)matrices[act_m_no] + i*dna_len +j)->prev_i; 
  int prev_j    = ((Pre_score*)matrices[act_m_no] + i*dna_len +j)->prev_j;
  int prev_m_no = ((Pre_score*)matrices[act_m_no] + i*dna_len +j)->prev_matrix_no;

  if(!(isfinite(((Pre_score*)matrices[act_m_no] + i*dna_len +j)->value))) {
    //cout << "Kein weiterer Pfad vorhanden\n";
    return 1;
  }

  if(((Pre_score*)matrices[act_m_no] + i*dna_len +j)->value==0) {
    //cout << "Kein weiterer Pfad vorhanden\n";
    return 1;
  }

  assert(dna_pos == j) ;
  assert(est_pos == i) ;
  splice_state = 0 ; //exon
  est_state = 1 ; //[112222111 ... for exons of length 2,4,3 ...

  //compute length of alignment (between max(m,n) and m+n)	
  //local_alignment: backtracking to first Zero: look at actual value
  while(!(((Pre_score*)matrices[act_m_no] + i*dna_len +j)->value==0)) {
    
    //just some info
    //if ((z>0) && (!(z==prev_m_no))) {
    //  fprintf(stdout, "(%d,%d): next step in next matrix\n", i,j) ;
    //}

    if ((prev_i == (i-1)) && (prev_j == (j-1))) { // match or mismatch
      (*result_length_ptr)= (*result_length_ptr) + 1;
      
      s_align[dna_pos-1] = splice_state; 
      dna_pos-- ;
      e_align[est_pos-1] = est_state ; //1 or 2, depended
      est_pos-- ;
      
      dnanum = check_char(dna[j-1]) ; //-1 because index starts with 0
      estnum = check_char(est[i-1]) ; //-1 because index starts with 0
      mparam[mlen*dnanum +estnum] ++ ;
    }
    else if ((prev_i == (i)) && (prev_j == (j-1))) {// gap on EST
      (*result_length_ptr)= (*result_length_ptr) + 1;
      
      s_align[dna_pos-1] = splice_state; 
      dna_pos-- ;

      dnanum = check_char(dna[j-1]) ; //-1 because index starts with 0
      estnum = 0 ; //gap 
      mparam[mlen*dnanum +estnum] ++ ;
    }
    else if ((prev_i == (i-1)) && (prev_j == (j))) { // gap on DNA
      (*result_length_ptr)= (*result_length_ptr) + 1;
      
      e_align[est_pos-1] = est_state ; //1 or 2, depended
      est_pos-- ;

      dnanum = 0 ; //gap
      estnum = check_char(est[i-1]) ; //-1 because index starts with 0
      mparam[mlen*dnanum +estnum] ++ ;
    }
    else {// splice site
      (*result_length_ptr) =  (*result_length_ptr) + (j - prev_j);
      if (((Pre_score*)matrices[act_m_no] + i*dna_len + j)->issplice==0) {
	fprintf(stdout, "matrices[act_m_no](i,j): %3.15f\n",((Pre_score*)matrices[act_m_no] + i*dna_len +j)->value) ;
	fprintf(stdout, "matrices[prev_m_no](i,j): %3.15f\n",((Pre_score*)matrices[prev_m_no] + i*dna_len +j)->value) ;
	fprintf(stdout, "issplice(act): %d\n",((Pre_score*)matrices[act_m_no] + i*dna_len + j)->issplice) ;
	fprintf(stdout, "issplice(prev): %d\n",((Pre_score*)matrices[prev_m_no] + i*dna_len + j)->issplice) ;
	fprintf(stdout, "act_m_no: %d, prev_m_no: %d\n", act_m_no, prev_m_no) ;
	fprintf(stdout, "j: %d -> prev_j:%d\n", j, prev_j) ;
	fprintf(stdout, "i: %d -> prev_i:%d\n", i, prev_i) ;
      }
      if (est_state==1) { //was exon labeled "1"
	est_state = 2 ; //new exon is labeled "2"
      }
      else {
	est_state = 1 ; //last exon was labeled "2", new exon is labeled "1"
      }
      for (; j > prev_j; j--) {
	if (splice_state == 0) { //coming from exon
	  splice_state = 2 ; //first intron_state: acceptor
	} else {
	  splice_state = 3 ; //intron
	}
	if (j-1 == prev_j) { //last intron_state: donor
	  splice_state = 1 ;//donor
	}
	s_align[dna_pos-1] = splice_state; 
	dna_pos-- ;
      }
      splice_state = 0 ; //exon again
      
    }
   
    startend[1] = i;
    startend[2] = j;
    i = prev_i;
    j = prev_j;
    act_m_no = prev_m_no ;
    
    prev_i    = ((Pre_score*)matrices[act_m_no] + i*dna_len + j)->prev_i; 
    prev_j    = ((Pre_score*)matrices[act_m_no] + i*dna_len + j)->prev_j; 
    prev_m_no = ((Pre_score*)matrices[act_m_no] + i*dna_len +j)->prev_matrix_no;
  }
  for (; dna_pos>0; dna_pos--) {
    s_align[dna_pos-1] = 4 ; // not aligned (-1 because array starts with index 0)
  }
  for (; est_pos>0; est_pos--) {
    e_align[est_pos-1] = 4 ; // not aligned (-1 because array starts with index 0)
  }

  //fprintf(stdout, "(%d, %d) --> (%d, %d)\n", startend[1], startend[2], startend[3], startend[4]) ;
  //fprintf(stdout, "result_length: %d\n", *result_length_ptr) ;


  return 0;
}


